@@ -1,16 +1,18 @@
 import datetime
 import decimal
-import getpass
+import pwinput
 import json
 import os
 import re
 import shutil
 import subprocess
 import sys
+import requests
+import time
 
 import defog
-from defog import Defog
-from defog.util import parse_update
+from defog.util import get_feedback, parse_update
+from prompt_toolkit import prompt
 
 USAGE_STRING = """
 Usage: defog <command>
@@ -18,12 +20,12 @@ Usage: defog <command>
 Available commands:
     init\t\t\tSetup defog credentials and your database connection
     gen <table1> <table2>\tSpecify tables to generate schema for
-    update <url>\t\tupdate schema (google sheets url) to defog
+    update <csv>\t\tupdate schema (path to CSV) to defog
     query\t\t\tRun a query
     deploy <gcp|aws>\t\tDeploy a defog server as a cloud function
-    check\t\t\tCheck if your database is suitable for defog (for prospective customers)
     quota\t\t\tCheck your API quota limits
     docs\t\t\tPrint documentation
+    serve\t\t\tServe a defog server locally
 """
 
 home_dir = os.path.expanduser("~")
@@ -43,13 +45,17 @@ def main():
         query()
     elif sys.argv[1] == "deploy":
         deploy()
-    elif sys.argv[1] == "check":
-        check_suitability()
     elif sys.argv[1] == "quota":
         quota()
     elif sys.argv[1] == "docs":
         # TODO
         raise NotImplementedError("docs not implemented yet")
+    elif sys.argv[1] == "serve":
+        serve()
+    elif sys.argv[1] == "serve-webserver":
+        serve_webserver()
+    elif sys.argv[1] == "serve-static":
+        serve_static()
     else:
         print(f"Unknown command: {sys.argv[1]}")
         print(USAGE_STRING)
@@ -68,14 +74,14 @@ def init():
     filepath = os.path.join(home_dir, ".defog", "connection.json")
     if os.path.exists(filepath):
         print(
-            "It looks like you've already initialized defog. Do you want to overwrite your existing configuration? (y/n)"
+            "It looks like you've already initialized defog. Do you want to re-enter your database credentials? (y/n)"
         )
-        overwrite = input()
+        overwrite = prompt().strip()
         if overwrite.lower() != "y":
             print("We'll keep your existing config. No changes were made.")
             sys.exit(0)
         else:
-            print("We'll overwrite a new config file at ~/.defog/connection.json")
+            print("We'll create a new config file at ~/.defog/connection.json")
     else:
         print("We'll create a new config file at ~/.defog/connection.json")
         if not os.path.exists(os.path.join(home_dir, ".defog")):
@@ -88,32 +94,41 @@ def init():
         )
         api_key = os.environ.get("DEFOG_API_KEY")
     else:
-        api_key = getpass.getpass(
-            prompt="Please enter your DEFOG_API_KEY. You can get it from https://defog.ai/account and creating an account:"
+        print(
+            "Please enter your DEFOG_API_KEY. You can get it from https://defog.ai/accounts/dashboard/ and creating an account:"
         )
+        api_key = prompt().strip()
+
     # prompt user for db_type
     print(
         "What database are you using? Available options are: "
         + ", ".join(defog.SUPPORTED_DB_TYPES)
     )
-    db_type = input().lower()
+    db_type = prompt().strip()
+    db_type = db_type.lower()
     while db_type not in defog.SUPPORTED_DB_TYPES:
         print(
             "Sorry, we don't support that database yet. Available options are: "
             + ", ".join(defog.SUPPORTED_DB_TYPES)
         )
-        db_type = input().lower()
+        db_type = prompt().strip()
+        db_type = db_type.lower()
     # depending on db_type, prompt user for appropriate db_creds
     if db_type == "postgres" or db_type == "redshift":
         print("Please enter your database host:")
-        host = input()
+        host = prompt().strip()
+        # check if host has http:// or https:// and remove it if it does
+        if host.startswith("http://") or host.startswith("https://"):
+            host = host.split("://")[1]
         print("Please enter your database port:")
-        port = input()
+        port = prompt().strip()
         print("Please enter your database name:")
-        database = input()
+        database = prompt().strip()
         print("Please enter your database user:")
-        user = input()
-        password = getpass.getpass(prompt="Please enter your database password:")
+        user = prompt().strip()
+        password = pwinput.pwinput(
+            prompt="Please enter your database password. This won't be displayed as you're typing for privacy reasons"
+        )
         db_creds = {
             "host": host,
             "port": port,
@@ -124,13 +139,13 @@ def init():
 
     elif db_type == "mysql":
         print("Please enter your database host:")
-        host = input()
+        host = prompt().strip()
         print("Please enter your database name:")
-        database = input()
+        database = prompt().strip()
         print("Please enter your database user:")
-        user = input()
+        user = prompt().strip()
         print("Please enter your database password:")
-        password = input()
+        password = pwinput.pwinput(prompt="Please enter your database password:")
         db_creds = {
             "host": host,
             "database": database,
@@ -139,28 +154,35 @@ def init():
         }
     elif db_type == "snowflake":
         print("Please enter your database account:")
-        account = input()
+        account = prompt().strip()
         print("Please enter your database warehouse:")
-        warehouse = input()
+        warehouse = prompt().strip()
         print("Please enter your database user:")
-        user = input()
-        print("Please enter your database password:")
-        password = input()
+        user = prompt().strip()
+        password = pwinput.pwinput(prompt="Please enter your database password:")
         db_creds = {
             "account": account,
             "warehouse": warehouse,
             "user": user,
             "password": password,
         }
-    elif db_type == "mongo" or db_type == "sqlserver":
-        print("Please enter your database connection string:")
-        connection_string = input()
+    elif db_type == "databricks":
+        print("Please enter your databricks host:")
+        host = prompt().strip()
+        token = pwinput.pwinput(prompt="Please enter your databricks token:")
+        print("Please add your http_path:")
+        http_path = prompt().strip()
+        print("Please enter your schema name (this is usually 'default'):")
+        schema = prompt().strip()
         db_creds = {
-            "connection_string": connection_string,
+            "server_hostname": host,
+            "access_token": token,
+            "http_path": http_path,
+            "schema": schema,
         }
     elif db_type == "bigquery":
         print("Please enter your bigquery json key's path:")
-        json_key_path = input()
+        json_key_path = prompt().strip()
         db_creds = {
             "json_key_path": json_key_path,
         }
@@ -178,18 +200,75 @@ def init():
     print(
         "Please enter the names of the tables you would like to register, separated by a space:"
     )
-    table_names = input()
+    table_names = prompt().strip()
     table_name_list = re.split(r"\s+", table_names.strip())
     # if input is empty, exit
     if table_name_list == [""]:
         print("No tables were registered. Exiting.")
         sys.exit(0)
     else:
+        print(
+            "Do you want to automatically scan these tables to determine which column might be categorical? The distinct values in each categorical column will be sent to our server. (y/n)"
+        )
+        scan_option = prompt().strip()
+        if scan_option.lower() == "y" or scan_option.lower() == "yes":
+            scan = True
+        else:
+            scan = False
         df = defog.Defog(api_key=api_key, db_type=db_type, db_creds=db_creds)
-        gsheets_url = df.generate_db_schema(table_name_list)
-        print("Your schema has been generated and is available at:\n")
-        print(f"\033[1m{gsheets_url}\033[0m.\n")
-        print("Next, please run `defog update <url>` to update the updated schema.")
+        filename = df.generate_db_schema(table_name_list, scan=scan)
+        pwd = os.getcwd()
+        print(
+            "Your schema has been generated and is available as a CSV file in this folder at:\n"
+        )
+        print(f"\033[1m{pwd}/{filename}\033[0m\n")
+
+    print(
+        "You can give us more context about your schema by editing the CSV above. Refer to our cookbook on how to do this here: https://defog.notion.site/Cookbook-for-schema-definitions-1650a6855ea447fdb0be75d39975571b#2ba1d37e243e4da3b8f17590b4a3e4e3.\n\nOnce you're done, you can just hit enter to upload the data in the CSV to Defog. If you would like to exit instead, just enter `exit`."
+    )
+    upload_option = prompt().strip()
+    if upload_option == "exit":
+        print("Exiting.")
+        sys.exit(0)
+    else:
+        print(
+            "We are now uploading this schema to Defog. This might take up to 30 seconds..."
+        )
+        resp = df.update_db_schema(filename)
+        if resp["status"] == "success":
+            print("Your schema has been updated. You're ready to start querying!")
+        else:
+            print("There was an error updating your schema:")
+            print(resp)
+            print("Please try again, or contact us at founders@defog.ai")
+            sys.exit(1)
+
+    print(
+        "You're all set! You can now start querying your database using `defog query`."
+    )
+
+    print(
+        f"You get started, try entering a sample question here, like how many rows are in the table {table_name_list[0]}?"
+    )
+
+    query = prompt("Enter your query here: ")
+    while query != "e":
+        resp = df.run_query(query, retries=3)
+        if not resp["ran_successfully"]:
+            print(f"Your query did not run successfully. Please try again.")
+        else:
+            print("Your question generated the following query:\n")
+            print(f"\033[1m{resp['query_generated']}\033[0m\n")
+            print("Results:\n")
+            # print results in tabular format using 'columns' and 'data' keys
+            try:
+                print_table(resp["columns"], resp["data"])
+            except:
+                print(resp)
+        query = prompt("Enter another query, or type 'e' to exit: ")
+
+    print("Exiting.")
+    sys.exit(0)
 
 
 def gen():
@@ -201,15 +280,35 @@ def gen():
         print(
             "defog gen requires a list of tables to generate. Please enter the names of the tables whose schema you would like to generate, separated by a space:"
         )
-        table_names = input()
+        table_names = prompt().strip()
         table_name_list = re.split(r"\s+", table_names.strip())
     else:
         table_name_list = sys.argv[2:]
-    gsheets_url = df.generate_db_schema(table_name_list)
-    print("Your schema has been generated and is available at:\n")
-    print(f"\033[1m{gsheets_url}\033[0m.\n")
+
+    if table_name_list == [""] or table_name_list == []:
+        print("No tables were registered. Exiting.")
+        sys.exit(0)
     print(
-        "If you do modify the schema in the link provided, please run `defog update <url>` to update the updated schema."
+        "Do you want to automatically scan these tables to determine which column might be categorical? The distinct values in each column likely to be a categorical column will be sent to our server. (y/n)"
+    )
+    scan_option = prompt().strip()
+    if scan_option.lower() == "y" or scan_option.lower() == "yes":
+        scan = True
+    else:
+        scan = False
+
+    filename = df.generate_db_schema(table_name_list, scan=scan)
+    pwd = os.getcwd()
+    print(
+        "Your schema has been generated and is available at the following CSV file in this folder:\n"
+    )
+    print(f"\033[1m{pwd}/{filename}\033[0m\n")
+
+    print("We are now uploading this auto-generated schema to Defog.")
+    df.update_db_schema(filename)
+
+    print(
+        "If you modify the auto-generated schema, please run `defog update <csv_filename>` again to refresh the schema on Defog's servers."
     )
 
 
@@ -220,15 +319,15 @@ def update():
     # check for 3rd arg (url), if not there, prompt user for url
     if len(sys.argv) < 3:
         print(
-            "defog update requires a google sheets url. Please enter the url of the google sheets document you would like to update:"
+            "defog update requires a CSV that contains your Database metadata. Please enter the path to the CSV you would like to update:"
         )
-        gsheets_url = input()
+        filename = prompt().strip()
     else:
-        gsheets_url = sys.argv[2]
+        filename = sys.argv[2]
     # load config from .defog/connection.json
     df = defog.Defog()
     # upload schema to defog
-    resp = df.update_db_schema(gsheets_url)
+    resp = df.update_db_schema(filename)
     if resp["status"] == "success":
         print("Your schema has been updated. You're ready to start querying!")
     else:
@@ -244,23 +343,57 @@ def query():
     df = defog.Defog()  # load config from .defog/connection.json
     if len(sys.argv) < 3:
         print("defog query requires a query. Please enter your query:")
-        query = input()
+        query = prompt().strip()
     else:
         query = sys.argv[2]
-    while query != "e":
+
+    user_question = ""
+    sql_generated = ""
+    while True:
+        if query == "e":
+            print("Exiting.")
+            sys.exit(0)
+        elif query == "":
+            print("Your query cannot be empty.")
+            query = prompt("Enter a query, or type 'e' to exit: ")
+        user_question = query
         resp = df.run_query(query, retries=3)
         if not resp["ran_successfully"]:
-            print(f"Your query did not run successfully. Please try again.")
+            if "query_generated" in resp:
+                print("Defog generated the following query to answer your question:\n")
+                print(f"\033[1m{resp['query_generated']}\033[0m\n")
+
+                print(
+                    f"However, your query did not run successfully. The error message generated while running the query on your database was\n\n\033[1m{resp['error_message']}\033[0m\n."
+                )
+
+                print(
+                    f"If you continue to get these errors, please consider updating the metadata in your schema by editing the google CSV generated and running `defog update <path/to/csv>`, or by updating your glossary.\n"
+                )
+            else:
+                print(
+                    f"Defog was unable to generate a query for your question. The error message generated while running the query on your database was\n\n\033[1m{resp.get('error_message')}\033[0m\n."
+                )
         else:
-            print("Your question generated the following query:\n")
+            sql_generated = resp.get("query_generated")
+            print("Defog generated the following query to answer your question:\n")
             print(f"\033[1m{resp['query_generated']}\033[0m\n")
+            reason_for_query = resp.get("reason_for_query", "")
+            if reason_for_query:
+                reason_for_query = reason_for_query.replace(". ", "\n\n")
+                print("This was its reasoning for generating this query:\n")
+                print(f"\033[1m{reason_for_query}\033[0m\n")
+
             print("Results:\n")
             # print results in tabular format using 'columns' and 'data' keys
             try:
                 print_table(resp["columns"], resp["data"])
             except:
                 print(resp)
-        query = input("Enter another query, or type 'e' to exit: ")
+
+            print()
+            get_feedback(df.api_key, df.db_type, user_question, sql_generated)
+        query = prompt("Please enter another query, or type 'e' to exit: ")
 
 
 def deploy():
@@ -270,10 +403,15 @@ def deploy():
     # check args for gcp or aws
     if len(sys.argv) < 3:
         print("defog deploy requires a cloud provider. Please enter 'gcp' or 'aws':")
-        cloud_provider = input().lower()
+        cloud_provider = prompt().strip().lower()
     else:
         cloud_provider = sys.argv[2].lower()
 
+    if len(sys.argv) >= 4:
+        function_name = sys.argv[3]
+    else:
+        function_name = f"defog-{cloud_provider}"
+
     # load config from .defog/connection.json
     df = defog.Defog()
 
@@ -285,7 +423,7 @@ def deploy():
             "gcloud",
             "functions",
             "deploy",
-            "defog-gcp",
+            function_name,
             "--runtime",
             "python310",
             "--region",
@@ -349,23 +487,6 @@ def deploy():
         raise ValueError("Cloud provider must be 'gcp' or 'aws'.")
 
 
-def check_suitability():
-    """
-    Check if tables in customer's schema are supported by our product.
-    """
-    df = defog.Defog()  # Note that api key can be invalid here for presales purposes
-    # parse list of tables from args
-    if len(sys.argv) < 3:
-        print(
-            "defog check-suitability requires a list of tables. Please enter your table names, separated by a space:"
-        )
-        table_names = input()
-        table_name_list = re.split(r"\s+", table_names.strip())
-    else:
-        table_name_list = sys.argv[2:]
-    df.check_db_suitability(tables=table_name_list)  # prints out messages to stdout
-
-
 def quota():
     """
     Check your current usage and quota.
@@ -378,7 +499,7 @@ def quota():
             print(f"Your current usage is {resp['queries_made']} queries.")
         else:
             print(
-                f"You are currently on the free plan with {100-resp['queries_made']} queries remaining for the month."
+                f"You are currently on the free plan with {1000-resp['queries_made']} queries remaining for the month."
             )
             print(f"Your current usage is {resp['queries_made']} queries.")
     else:
@@ -439,5 +560,54 @@ def print_table(columns, data):
         print()
 
 
+def serve_webserver():
+    from defog.serve import app
+    import uvicorn
+
+    uvicorn.run(app, host="localhost", port=1235)
+
+
+def serve_static():
+    import http.server
+    import socketserver
+    import webbrowser
+
+    port = 8002
+    directory = os.path.join(defog.__path__[0], "static")
+
+    class Handler(http.server.SimpleHTTPRequestHandler):
+        def __init__(self, *args, **kwargs):
+            super().__init__(*args, directory=directory, **kwargs)
+
+    webbrowser.open(f"http://localhost:{port}")
+    with socketserver.TCPServer(("", port), Handler) as httpd:
+        print(f"Serving at port {port}")
+        print(f"Static folder is {directory}")
+        httpd.serve_forever()
+
+
+def serve():
+    """
+    Serve a defog server locally.
+    """
+    print("Starting defog server...")
+    print("Serving webserver...")
+    webserver_process = subprocess.Popen(["defog", "serve-webserver"])
+    time.sleep(2)
+
+    print("Serving static files...")
+    static_process = subprocess.Popen(["defog", "serve-static"])
+
+    print("Press Ctrl+C to exit.")
+    try:
+        while True:
+            pass
+    except KeyboardInterrupt:
+        print("Exiting...")
+        static_process.terminate()
+        webserver_process.terminate()
+        sys.exit(0)
+
+
 if __name__ == "__main__":
     main()
